package nxt.tools;

import nxt.Constants;
import nxt.Nxt;
import nxt.NxtException;
import nxt.addons.JA;
import nxt.addons.JO;
import nxt.blockchain.ChainTransactionId;
import nxt.blockchain.ChildChain;
import nxt.http.API;
import nxt.http.callers.GetContractReferencesCall;
import nxt.lightcontracts.ContractReferenceAttachment;
import nxt.lightcontracts.ContractReferenceDeleteAttachment;
import nxt.taggeddata.TaggedDataAttachment;
import nxt.util.Convert;
import nxt.util.ResourceLookup;
import nxt.util.Search;
import nxt.util.TrustAllSSLProvider;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.DefaultParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;

import javax.net.ssl.HttpsURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

public class ContractManager {

    private static final String CLI_HEADER = "Use Contract Manager to Upload or Reference Lightweight Contracts";

    private String secretPhrase;
    private long feeNQT;
    private long feeRateNQTPerFXT;
    private long minBundlerBalanceFXT;
    private ChildChain childChain;
    private JO contractSetup;

    enum OPTION {
        NAME('n', "name", true, "contract name", false, (OPTION)null),
        HASH('h', "hash", true, "contract full hash", false, (OPTION)null),
        ACCOUNT('a', "account", true, "account id", false, (OPTION)null),
        UPLOAD('u', "upload", false, "upload new contract", true, NAME),
        REFERENCE('r', "reference", false, "reference existing contract", true, HASH, NAME),
        DELETE('d', "delete", false, "delete reference", true, ACCOUNT, NAME),
        LIST('l', "list", false, "list contract references for account", true, ACCOUNT);

        private final char opt;
        private final String longOpt;
        private boolean hasArgs;
        private String description;
        private boolean isAction;
        private OPTION[] dependencies;

        OPTION(char opt, String longOpt, boolean hasArgs, String description, boolean isAction, OPTION... dependencies) {
            this.opt = opt;
            this.longOpt = longOpt;
            this.hasArgs = hasArgs;
            this.description = description;
            this.isAction = isAction;
            this.dependencies = dependencies;
        }

        public String getOpt() {
            return Character.toString(opt);
        }

        public String getLongOpt() {
            return longOpt;
        }

        public boolean hasArgs() {
            return hasArgs;
        }

        public String getDescription() {
            return description;
        }

        public boolean isAction() {
            return isAction;
        }

        public OPTION[] getDependencies() {
            return dependencies;
        }
    }

    public static void main(String[] args) {
        ContractManager contractManager = new ContractManager();
        Options options = new Options();
        Arrays.stream(OPTION.values()).forEach(o -> options.addOption(new Option(o.getOpt(), o.getLongOpt(), o.hasArgs(), o.getDescription())));

        CommandLineParser parser = new DefaultParser();
        HelpFormatter formatter = new HelpFormatter();
        CommandLine cmd;
        try {
            cmd = parser.parse(options, args);
        } catch (ParseException e) {
            System.out.println(e.getMessage());
            formatter.printHelp(ContractManager.class.getName(), options);
            return;
        }
        List<OPTION> specifiedOptions = Arrays.stream(OPTION.values()).filter(o -> cmd.hasOption(o.getOpt())).collect(Collectors.toList());
        List<OPTION> actionOptions = specifiedOptions.stream().filter(OPTION::isAction).collect(Collectors.toList());
        if (actionOptions.size() != 1) {
            String actions = Arrays.stream(OPTION.values()).filter(OPTION::isAction).map(o -> "--" + o.getLongOpt()).collect(Collectors.joining(","));
            formatter.printHelp(ContractManager.class.getName(), CLI_HEADER, options, String.format("Exactly one of the arguments %s has to be specified", actions));
            return;
        }
        OPTION action = actionOptions.get(0);
        OPTION[] dependencies = action.getDependencies();
        String missingDependencies = Arrays.stream(dependencies).filter(d -> !specifiedOptions.contains(d)).map(o -> "--" + o.getLongOpt()).collect(Collectors.joining(","));
        if (missingDependencies.length() > 0) {
            formatter.printHelp(ContractManager.class.getName(), CLI_HEADER, options, String.format("Action --%s is missing the following arguments %s", action.getLongOpt(), missingDependencies));
            return;
        }

        if (action == OPTION.LIST) {
            if (!cmd.hasOption("account")) {
                formatter.printHelp(ContractManager.class.getName(), CLI_HEADER, options, "Specify an Ardor account for which to list the contract reference");
                return;
            }
            contractManager.list(cmd.getOptionValue("account"), cmd.getOptionValue("name"));
            return;
        }
        String contractName = cmd.getOptionValue("name");
        contractManager.init(contractName);
        if (action == OPTION.UPLOAD) {
            byte[] contractFullHash = contractManager.upload(contractName);
            contractManager.reference(contractName, contractFullHash);
        } else if (action == OPTION.REFERENCE) {
            if (!cmd.hasOption("h")) {
                formatter.printHelp(ContractManager.class.getName(), CLI_HEADER, options, "Specify contract full hash");
                return;
            }
            byte[] contractFullHash = Convert.parseHexString(cmd.getOptionValue("hash"));
            contractManager.reference(contractName, contractFullHash);
        } else if (action == OPTION.DELETE) {
            if (!cmd.hasOption("account")) {
                formatter.printHelp(ContractManager.class.getName(), CLI_HEADER, options, "Specify an Ardor account for which to delete the contract reference");
                return;
            }
            contractManager.delete(cmd.getOptionValue("account"), contractName);
        } else {
            formatter.printHelp(ContractManager.class.getName(), CLI_HEADER, options, "No action specified");
        }
    }

    public void init(String contractName) {
        secretPhrase = Convert.emptyToNull(Nxt.getStringProperty("contract.manager.secretPhrase", null, true));
        if (secretPhrase == null) {
            throw new IllegalArgumentException("contract.manager.secretPhrase not specified");
        }
        feeNQT = Nxt.getIntProperty("contract.manager.feeNQT", -1);
        feeRateNQTPerFXT = Nxt.getIntProperty("contract.manager.feeRateNQTPerFXT", -1);
        minBundlerBalanceFXT = Nxt.getIntProperty("contract.manager.minBundlerBalanceFXT", 0);
        childChain = ChildChain.IGNIS;
        String contractUploadParamsFile = Nxt.getStringProperty("contract.manager.uploadParamsFile", "contract.uploader.json");
        JO contractUploaderConfig = ResourceLookup.loadJsonResource(contractUploadParamsFile);
        if (contractUploaderConfig == null) {
            return;
        }
        JA contracts = contractUploaderConfig.getArray("contracts");
        List<JO> contractList = contracts.objects();
        contractSetup = contractList.stream().filter(c -> contractName.equals(c.getString("className"))).findFirst().orElse(null);
        if (contractSetup == null) {
            throw new IllegalArgumentException(String.format("Contract definition for %s not found in %s", contractName, contractUploadParamsFile));
        }
    }

    public void list(String account, String name) {
        JO response = listImpl(account, name);
        if (response.isExist("contractReferences")) {
            List<JO> references = response.getJoList("contractReferences");
            for (JO reference : references) {
                System.out.println(reference.toJSONString());
            }
        } else {
            System.out.println("No contract references found");
        }
    }

    public JO listImpl(String account, String name) {
        GetContractReferencesCall getContractReferences = GetContractReferencesCall.create().account(account);
        if (name != null) {
            getContractReferences.contractName(name);
        }
        getContractReferences.remote(getUrl());
        return getContractReferences.call();
    }

    public byte[] upload(String contractName) {
        JO response = uploadImpl(contractName);
        if (!response.isExist("fullHash")) {
            return null;
        }
        return response.parseHexString("fullHash");
    }

    public JO uploadImpl(String contractName) {
        String packageName = contractSetup.getString("packageName");
        String filePath = contractSetup.getString("filePath");
        if (filePath == null) {
            filePath = (packageName + "." + contractName).replace('.', '/') + ".class";
        }
        String fileName = filePath.substring(filePath.lastIndexOf('/') + 1);
        byte[] resourceBytes = ResourceLookup.getResourceBytes(filePath);
        if (resourceBytes == null) {
            throw new IllegalStateException("Cannot load " + filePath);
        }
        String type;
        String detectedMimeType = Search.detectMimeType(resourceBytes, fileName);
        if (detectedMimeType != null) {
            type = detectedMimeType.substring(0, Math.min(detectedMimeType.length(), Constants.MAX_TAGGED_DATA_TYPE_LENGTH));
        } else {
            throw new IllegalStateException("Cannot detect file type " + filePath);
        }
        TaggedDataAttachment attachment;
        try {
            JO contractDescription = new JO();
            if (contractSetup.isExist("version")) {
                contractDescription.put("version", contractSetup.getString("version"));
            }
            if (contractSetup.isExist("url")) {
                contractDescription.put("url", contractSetup.getString("url"));
            }
            if (contractSetup.isExist("comment")) {
                contractDescription.put("comment", contractSetup.getString("comment"));
            }
            attachment = new TaggedDataAttachment(packageName + "." + contractName,
                    contractDescription.toJSONString(), contractSetup.getString("tags", ""), type, "contracts", false, fileName, resourceBytes);
        } catch (NxtException.NotValidException e) {
            throw new IllegalArgumentException(e);
        }
        JO uploadContractTransaction = LocalSigner.signAndBroadcast(childChain, 0, attachment, secretPhrase, feeNQT, feeRateNQTPerFXT, minBundlerBalanceFXT, null, getUrl());
        if (!uploadContractTransaction.isExist("fullHash")) {
            System.out.printf("Contract not uploaded %s\n", contractName);
            return uploadContractTransaction;
        }
        System.out.printf("Contract class %s uploaded %s\n", attachment.getName(), uploadContractTransaction.getString("fullHash"));
        return uploadContractTransaction;
    }

    public void reference(String contractName, byte[] contractFullHash) {
        reference(contractName, contractFullHash, null);
    }

    public JO reference(String contractName, byte[] contractFullHash, JO contractParams) {
        ChainTransactionId contractId = new ChainTransactionId(childChain.getId(), contractFullHash);
        if (contractParams == null) {
            contractParams = contractSetup.getJo("params");
        }
        ContractReferenceAttachment attachment = new ContractReferenceAttachment(contractName, contractParams.toJSONString(), contractId);
        URL url = getUrl();
        JO contractReferenceTransaction = LocalSigner.signAndBroadcast(ChildChain.IGNIS, 0, attachment, secretPhrase, feeNQT, feeRateNQTPerFXT, minBundlerBalanceFXT, null, url);
        if (!contractReferenceTransaction.isExist("fullHash")) {
            System.out.printf("Contract reference not registered %s\n", contractName);
            return contractReferenceTransaction;
        }
        System.out.printf("Contract reference registered %s=%s for contract %s\n", attachment.getContractName(), attachment.getContractParams(), attachment.getContractId().toString());
        return contractReferenceTransaction;
    }

    public void delete(String account, String contractName) {
        JO response = listImpl(account, contractName);
        List<JO> contractReferences = response.getJoList("contractReferences");
        if (contractReferences.size() == 0) {
            System.out.printf("Cannot find contract reference for account %s contract name %s\n", account, contractName);
            return;
        }
        long referenceId = contractReferences.get(0).getEntityId("id");
        ContractReferenceDeleteAttachment attachment = new ContractReferenceDeleteAttachment(referenceId);
        URL url = getUrl();
        JO contractReferenceDeleteTransaction = LocalSigner.signAndBroadcast(ChildChain.IGNIS, 0, attachment, secretPhrase, feeNQT, feeRateNQTPerFXT, minBundlerBalanceFXT, null, url);
        if (!contractReferenceDeleteTransaction.isExist("fullHash")) {
            System.out.printf("Contract reference delete failed %s\n", contractReferenceDeleteTransaction.toJSONString());
            return;
        }
        System.out.printf("Contract reference %s deleted for account %s contract name %s\n", Long.toUnsignedString(attachment.getContractReferenceId()), account, contractName);
    }

    private URL getUrl() {
        String host = Convert.emptyToNull(Nxt.getStringProperty("contract.manager.serverAddress"));
        if (host == null) {
            return null;
        }
        String protocol = "http";
        boolean useHttps = Nxt.getBooleanProperty("contract.manager.useHttps");
        if (useHttps) {
            protocol = "https";
            HttpsURLConnection.setDefaultSSLSocketFactory(TrustAllSSLProvider.getSslSocketFactory());
            HttpsURLConnection.setDefaultHostnameVerifier(TrustAllSSLProvider.getHostNameVerifier());
        }
        int port = Constants.isTestnet ? API.TESTNET_API_PORT : Nxt.getIntProperty("nxt.apiServerPort");
        URL url;
        try {
            url = new URL(protocol, host, port, "/nxt");
        } catch (MalformedURLException e) {
            throw new IllegalStateException(e);
        }
        System.out.println("Connecting to URL " + url);
        return url;
    }

}
